@@ -52,6 +52,20 @@
 // 10 seconds
 #define NM_REQUESTSCAN_LIMIT_RATE 10000
 
+static const QString KEY_802_11_WIRELESS_SECURITY = QStringLiteral("802-11-wireless-security");
+
+const QString escapeSpecialCharacters(const QString &text)
+{
+    // See https://github.com/zxing/zxing/wiki/Barcode-Contents
+    QString result = text;
+    result.replace(u'\\', QStringLiteral("\\\\"))
+        .replace(u';', QStringLiteral("\\;"))
+        .replace(u',', QStringLiteral("\\,"))
+        .replace(u'"', QStringLiteral("\\\""))
+        .replace(u':', QStringLiteral("\\:"));
+    return result;
+}
+
 Handler::Handler(QObject *parent)
     : QObject(parent)
     , m_tmpWirelessEnabled(NetworkManager::isWirelessEnabled())
@@ -68,13 +82,22 @@ Handler::Handler(QObject *parent)
         NetworkManager::ActiveConnection::Ptr hotspot = NetworkManager::findActiveConnection(Configuration::self().hotspotConnectionPath());
         if (!hotspot) {
             Configuration::self().setHotspotConnectionPath(QString());
+            m_hotspotActive = false;
+            Q_EMIT hotspotActiveChanged(m_hotspotActive);
+        } else {
+            m_hotspotActive = true;
+            Q_EMIT hotspotActiveChanged(m_hotspotActive);
         }
+    } else {
+        m_hotspotActive = false;
+        Q_EMIT hotspotActiveChanged(m_hotspotActive);
     }
 
     m_hotspotSupported = checkHotspotSupported();
 
     if (NetworkManager::checkVersion(1, 16, 0)) {
         connect(NetworkManager::notifier(), &NetworkManager::Notifier::primaryConnectionTypeChanged, this, &Handler::primaryConnectionTypeChanged);
+        connect(NetworkManager::notifier(), &NetworkManager::Notifier::wirelessEnabledChanged, this, &Handler::wirelessEnabledChanged);
     }
 }
 
@@ -85,7 +108,7 @@ void Handler::activateConnection(const QString &connection, const QString &devic
     activateConnectionInternal(connection, device, specificObject);
 }
 
-QCoro::Task<void> Handler::activateConnectionInternal(const QString &connection, const QString &device, const QString &specificObject)
+QCoro::Task<> Handler::activateConnectionInternal(const QString &connection, const QString &device, const QString &specificObject)
 {
     NetworkManager::Connection::Ptr con = NetworkManager::findConnection(connection);
 
@@ -205,7 +228,7 @@ void Handler::requestWifiCode(const QString &connectionPath, const QString &ssid
 
     auto securityType = static_cast<NetworkManager::WirelessSecurityType>(_securityType);
 
-    QString ret = QStringLiteral("WIFI:S:") + ssid + QLatin1Char(';');
+    QString ret = QStringLiteral("WIFI:S:") + escapeSpecialCharacters(ssid) + QLatin1Char(';');
     if (securityType != NetworkManager::NoneSecurity) {
         switch (securityType) {
         case NetworkManager::NoneSecurity:
@@ -216,10 +239,8 @@ void Handler::requestWifiCode(const QString &connectionPath, const QString &ssid
             break;
         case NetworkManager::WpaPsk:
         case NetworkManager::Wpa2Psk:
-            ret += QStringLiteral("T:WPA;");
-            break;
         case NetworkManager::SAE:
-            ret += QStringLiteral("T:SAE;");
+            ret += QStringLiteral("T:WPA;");
             break;
         default:
         case NetworkManager::DynamicWep:
@@ -238,14 +259,12 @@ void Handler::requestWifiCode(const QString &connectionPath, const QString &ssid
         return;
     }
 
-    const auto key = QStringLiteral("802-11-wireless-security");
-    auto reply = connection->secrets(key);
+    auto reply = connection->secrets(KEY_802_11_WIRELESS_SECURITY);
     m_requestWifiCodeWatcher = new QDBusPendingCallWatcher(reply, this);
-    m_requestWifiCodeWatcher->setProperty("key", key);
-    m_requestWifiCodeWatcher->setProperty("ret", ret);
-    m_requestWifiCodeWatcher->setProperty("securityType", static_cast<int>(securityType));
-    m_requestWifiCodeWatcher->setProperty("ssid", ssid);
-    connect(m_requestWifiCodeWatcher, &QDBusPendingCallWatcher::finished, this, &Handler::slotRequestWifiCode);
+    connect(m_requestWifiCodeWatcher, &QDBusPendingCallWatcher::finished, this, [=, this](QDBusPendingCallWatcher *watcher) mutable {
+        // ret needs to stay mutable, so move it.  But ssid can be const.
+        onWiFiCodeResponse(watcher, std::move(ret), ssid, securityType);
+    });
 }
 
 void Handler::addAndActivateConnection(const QString &device, const QString &specificParameter, const QString &password)
@@ -253,7 +272,7 @@ void Handler::addAndActivateConnection(const QString &device, const QString &spe
     addAndActivateConnectionInternal(device, specificParameter, password);
 }
 
-QCoro::Task<void> Handler::addAndActivateConnectionInternal(const QString &device, const QString &specificObject, const QString &password)
+QCoro::Task<> Handler::addAndActivateConnectionInternal(const QString &device, const QString &specificObject, const QString &password)
 {
     NetworkManager::AccessPoint::Ptr ap;
     NetworkManager::WirelessDevice::Ptr wifiDev;
@@ -355,7 +374,7 @@ QCoro::Task<void> Handler::addAndActivateConnectionInternal(const QString &devic
     settings.clear();
 }
 
-QCoro::Task<void> Handler::addConnection(const NMVariantMapMap &map)
+QCoro::Task<> Handler::addConnection(const NMVariantMapMap &map)
 {
     const QString connectionId = map.value(QStringLiteral("connection")).value(QStringLiteral("id")).toString();
 
@@ -433,7 +452,7 @@ void Handler::addConnection(NMConnection *connection)
                               userData);
 }
 
-QCoro::Task<void> Handler::addAndActivateConnectionDBus(const NMVariantMapMap &map, const QString &device, const QString &specificObject)
+QCoro::Task<> Handler::addAndActivateConnectionDBus(const NMVariantMapMap &map, const QString &device, const QString &specificObject)
 {
     const QString name = map.value(QStringLiteral("connection")).value(QStringLiteral("id")).toString();
     QPointer<Handler> alive(this);
@@ -457,7 +476,7 @@ void Handler::deactivateConnection(const QString &connection, const QString &dev
     deactivateConnectionInternal(connection, device);
 }
 
-QCoro::Task<void> Handler::deactivateConnectionInternal(const QString &_connection, const QString &device)
+QCoro::Task<> Handler::deactivateConnectionInternal(const QString &_connection, const QString &device)
 {
     const QString connection = _connection;
     NetworkManager::Connection::Ptr con = NetworkManager::findConnection(connection);
@@ -530,7 +549,7 @@ void setBluetoothEnabled(const QString &path, bool enabled)
     QDBusConnection::systemBus().asyncCall(message);
 }
 
-QCoro::Task<void> Handler::enableBluetooth(bool enable)
+QCoro::Task<> Handler::enableBluetooth(bool enable)
 {
     qDBusRegisterMetaType<QMap<QDBusObjectPath, NMVariantMapMap>>();
 
@@ -581,6 +600,10 @@ void Handler::enableNetworking(bool enable)
 void Handler::enableWireless(bool enable)
 {
     NetworkManager::setWirelessEnabled(enable);
+    if (!enable && m_hotspotActive) {
+        m_hotspotActive = false;
+        Q_EMIT hotspotActiveChanged(m_hotspotActive);
+    }
 }
 
 void Handler::enableWwan(bool enable)
@@ -593,7 +616,7 @@ void Handler::removeConnection(const QString &connection)
     removeConnectionInternal(connection);
 }
 
-QCoro::Task<void> Handler::removeConnectionInternal(const QString &connection)
+QCoro::Task<> Handler::removeConnectionInternal(const QString &connection)
 {
     NetworkManager::Connection::Ptr con = NetworkManager::findConnection(connection);
 
@@ -633,7 +656,7 @@ QCoro::Task<void> Handler::removeConnectionInternal(const QString &connection)
     }
 }
 
-QCoro::Task<void> Handler::updateConnection(NetworkManager::Connection::Ptr connection, const NMVariantMapMap &map)
+QCoro::Task<> Handler::updateConnection(NetworkManager::Connection::Ptr connection, const NMVariantMapMap &map)
 {
     QPointer<Handler> thisGuard(this);
     QDBusReply<void> reply = co_await connection->update(map);
@@ -665,7 +688,7 @@ void Handler::requestScan(const QString &interface)
     requestScanInternal(interface);
 }
 
-QCoro::Task<void> Handler::requestScanInternal(const QString &interface)
+QCoro::Task<> Handler::requestScanInternal(const QString &interface)
 {
     for (const NetworkManager::Device::Ptr &device : NetworkManager::networkInterfaces()) {
         if (device->type() == NetworkManager::Device::Wifi) {
@@ -748,7 +771,7 @@ void Handler::createHotspot()
     createHotspotInternal();
 }
 
-QCoro::Task<void> Handler::createHotspotInternal()
+QCoro::Task<> Handler::createHotspotInternal()
 {
     bool foundInactive = false;
     bool useApMode = false;
@@ -858,11 +881,13 @@ QCoro::Task<void> Handler::createHotspotInternal()
         connect(hotspot.data(), &NetworkManager::ActiveConnection::stateChanged, [this](NetworkManager::ActiveConnection::State state) {
             if (state > NetworkManager::ActiveConnection::Activated) {
                 Configuration::self().setHotspotConnectionPath(QString());
-                Q_EMIT hotspotDisabled();
+                m_hotspotActive = false;
+                Q_EMIT hotspotActiveChanged(m_hotspotActive);
             }
         });
 
-        Q_EMIT hotspotCreated();
+        m_hotspotActive = true;
+        Q_EMIT hotspotActiveChanged(m_hotspotActive);
     }
 }
 
@@ -883,7 +908,8 @@ void Handler::stopHotspot()
     NetworkManager::deactivateConnection(activeConnectionPath);
     Configuration::self().setHotspotConnectionPath(QString());
 
-    Q_EMIT hotspotDisabled();
+    m_hotspotActive = false;
+    Q_EMIT hotspotActiveChanged(m_hotspotActive);
 }
 
 bool Handler::checkRequestScanRateLimit(const NetworkManager::WirelessDevice::Ptr &wifiDevice)
@@ -916,7 +942,7 @@ bool Handler::checkHotspotSupported()
                 wifiFound = true;
 
                 NetworkManager::WirelessDevice::Ptr wifiDev = device.objectCast<NetworkManager::WirelessDevice>();
-                if (wifiDev && !wifiDev->isActive()) {
+                if (wifiDev && wifiDev->state() == NetworkManager::WirelessDevice::Disconnected) {
                     unusedWifiFound = true;
                 }
             }
@@ -931,7 +957,7 @@ bool Handler::checkHotspotSupported()
         }
 
         // Check if the primary connection which is used for internet connectivity is not using WiFi
-        if (NetworkManager::primaryConnectionType() != NetworkManager::ConnectionSettings::Wireless) {
+        if (NetworkManager::primaryConnectionType() != NetworkManager::ConnectionSettings::Wireless && NetworkManager::isWirelessEnabled()) {
             return true;
         }
     }
@@ -984,6 +1010,13 @@ void Handler::primaryConnectionTypeChanged(NetworkManager::ConnectionSettings::C
     Q_EMIT hotspotSupportedChanged(m_hotspotSupported);
 }
 
+void Handler::wirelessEnabledChanged(bool wirelessEnabled)
+{
+    Q_UNUSED(wirelessEnabled)
+    m_hotspotSupported = checkHotspotSupported();
+    Q_EMIT hotspotSupportedChanged(m_hotspotSupported);
+}
+
 void Handler::unlockRequiredChanged(MMModemLock modemLock)
 {
     if (modemLock == MM_MODEM_LOCK_NONE) {
@@ -991,21 +1024,18 @@ void Handler::unlockRequiredChanged(MMModemLock modemLock)
     }
 }
 
-void Handler::slotRequestWifiCode(QDBusPendingCallWatcher *watcher)
+void Handler::onWiFiCodeResponse(QDBusPendingCallWatcher *watcher, QString &&ret, const QString &ssid, NetworkManager::WirelessSecurityType securityType)
 {
     watcher->deleteLater();
-
-    QString ret = watcher->property("ret").toString();
-    const QString ssid = watcher->property("ssid").toString();
     QDBusPendingReply<NMVariantMapMap> reply = *watcher;
     if (!reply.isValid() || reply.isError()) {
         Q_EMIT wifiCodeReceived(ret % QLatin1Char(';'), ssid);
         return;
     }
 
-    const auto secret = reply.argumentAt<0>()[watcher->property("key").toString()];
+    const auto secret = reply.argumentAt<0>()[KEY_802_11_WIRELESS_SECURITY];
     QString pass;
-    switch (static_cast<NetworkManager::WirelessSecurityType>(watcher->property("securityType").toInt())) {
+    switch (securityType) {
     case NetworkManager::NoneSecurity:
     case NetworkManager::OWE:
         break;
@@ -1019,10 +1049,10 @@ void Handler::slotRequestWifiCode(QDBusPendingCallWatcher *watcher)
         return;
     }
     if (!pass.isEmpty()) {
-        ret += QStringLiteral("P:") % pass % QLatin1Char(';');
+        ret += QStringLiteral("P:") % escapeSpecialCharacters(pass) % QLatin1Char(';');
     }
 
-    Q_EMIT wifiCodeReceived(ret % QLatin1Char(';'), ssid);
+    Q_EMIT wifiCodeReceived(ret % QLatin1Char(';'), ssid, pass);
 }
 
 #include "moc_handler.cpp"
